/*
 * CPM (clock and power management) support for the JZ4740.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <l4/devices/clocks.h>
#include <l4/sys/types.h>
#include <stdint.h>



#ifdef __cplusplus

#include <l4/devices/cpm.h>
#include <l4/devices/hw_register_block.h>

/* A simple abstraction for accessing the CPM registers.
 * A proper device could inherit from Hw::Device and use an
 * Int_property for _exclk_freq. */

class Cpm_jz4740_chip : public Cpm_chip_base
{
private:
  Hw::Register_block<32> _regs;
  uint32_t _exclk_freq;

  // Utility methods.

  uint8_t _get_divider(uint32_t reg, uint32_t mask, uint8_t shift);

  // PLL status.

  int have_pll();
  int pll_enabled();
  int pll_bypassed();

  // Clock control.

  uint32_t get_clock_gate_value(enum Clock_identifiers clock);

public:
  Cpm_jz4740_chip(l4_addr_t addr, uint32_t exclk_freq);

  // PLL configuration.

  uint16_t get_multiplier();
  uint8_t get_input_division();
  uint8_t get_output_division();

  // Divider configuration.

  uint8_t get_cpu_divider();
  uint8_t get_hclock_divider();
  uint8_t get_pclock_divider();
  uint8_t get_memory_divider();
  uint8_t get_source_divider();
  uint16_t get_lcd_pixel_divider();

  void set_lcd_device_divider(uint8_t division);
  void set_lcd_pixel_divider(uint16_t division);

  // PLL frequency status.

  uint32_t get_pll_frequency();
  uint32_t get_source_frequency();

  // Clock frequency status.

  uint32_t get_cpu_frequency();
  uint32_t get_hclock_frequency();
  uint32_t get_pclock_frequency();
  uint32_t get_memory_frequency();

  // Clock configuration.

  uint32_t get_frequency(enum Clock_identifiers clock);
  void set_frequency(enum Clock_identifiers clock, uint32_t frequency);

  int have_clock(enum Clock_identifiers clock);
  void start_clock(enum Clock_identifiers clock);
  void stop_clock(enum Clock_identifiers clock);

};

#endif /* __cplusplus */



/* C language interface. */

EXTERN_C_BEGIN

void *jz4740_cpm_init(l4_addr_t cpm_base);

int jz4740_cpm_have_pll(void *cpm);

int jz4740_cpm_have_clock(void *cpm, enum Clock_identifiers clock);
void jz4740_cpm_start_clock(void *cpm, enum Clock_identifiers clock);
void jz4740_cpm_stop_clock(void *cpm, enum Clock_identifiers clock);

uint16_t jz4740_cpm_get_lcd_pixel_divider(void *cpm);

uint32_t jz4740_cpm_get_frequency(void *cpm, enum Clock_identifiers clock);
void jz4740_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint32_t frequency);

uint32_t jz4740_cpm_get_cpu_frequency(void *cpm);
uint32_t jz4740_cpm_get_hclock_frequency(void *cpm);
uint32_t jz4740_cpm_get_source_frequency(void *cpm);
uint32_t jz4740_cpm_get_pclock_frequency(void *cpm);
uint32_t jz4740_cpm_get_memory_frequency(void *cpm);

EXTERN_C_END
